from graphql import (
    GraphQLArgument,
    GraphQLField,
    GraphQLInputField,
    GraphQLInputObjectType,
    GraphQLList,
    GraphQLObjectType,
    GraphQLSchema,
    GraphQLNonNull,
)
from sqlalchemy import inspect

from .errors import ConfigurationError
from .filters import convert_to_filter
from .models import Table
from .resolver import build_async_resolver, build_sync_resolver
from .scalars import convert_to_scalar, IntScalar, OrderingEnumScalar


def _validate_table(table: Table, inspected):
    """
    Perform validations on the table registered. Validations include:
     - Checking all fields exist
     - Checking all types are supported

    Any failed validations causes an exception to be thrown.
    """
    # Check all fields exist
    for field in (
        set(table.include_fields or [])
        | set(table.exclude_fields)
        | set(table.filter_fields)
        | set(table.order_fields)
        | (set(table.default_order.keys()) if table.default_order else set())
    ):
        if field not in inspected.columns:
            raise ConfigurationError(
                f"Field {field} does not exist in {table.graphql_name}!"
            )


def build_gql_schema(tables: list[Table], is_async: bool) -> GraphQLSchema:
    # Mapping of types to scalars
    scalar_map = {}

    # Step 1 - Build list of table schemas
    table_schemas: dict[str, GraphQLField] = {}

    for table in tables:
        inspected = inspect(table.sqlalchemy_cls)

        _validate_table(table, inspected)

        fields = {}
        query_fields = {}

        for col in inspected.columns:
            if col.type.python_type in scalar_map:
                ql_type = scalar_map[col.type.python_type]
            else:
                ql_type = convert_to_scalar(col)
                scalar_map[col.type.python_type] = ql_type

            if col.key not in table.exclude_fields and (
                not table.include_fields or col.key in table.include_fields
            ):
                if col.nullable:
                    fields[col.key] = GraphQLField(ql_type)  # type: ignore
                else:
                    fields[col.key] = GraphQLField(GraphQLNonNull(ql_type))  # type: ignore

            if col.key in table.filter_fields:
                query_fields[col.key] = convert_to_filter(col, ql_type)

        args = {}

        if query_fields:
            args["filter"] = GraphQLArgument(
                GraphQLInputObjectType(
                    name=f"{table.graphql_name}_filter", fields=lambda: query_fields
                )  # type: ignore
            )

        if table.pagination:
            args["limit"] = GraphQLArgument(
                IntScalar, default_value=table.default_limit
            )
            args["offset"] = GraphQLArgument(IntScalar, default_value=0)

        if table.order_fields:
            f = {it: GraphQLInputField(OrderingEnumScalar) for it in table.order_fields}
            args["order"] = GraphQLArgument(
                GraphQLInputObjectType(
                    name=f"{table.graphql_name}_order", fields=lambda: f
                )  # type: ignore
            )

        ql_table = GraphQLObjectType(
            name=table.graphql_name, description=table.description, fields=fields
        )

        table_schemas[table.graphql_name + "s"] = GraphQLField(
            GraphQLList(ql_table),
            resolve=build_async_resolver(table)
            if is_async
            else build_sync_resolver(table),
            args=args,
        )

    query = GraphQLObjectType(name="Query", fields=lambda: table_schemas)
    return GraphQLSchema(query=query)  # type: ignore



from abc import ABC
from typing import Any
from graphql import ExecutionResult, GraphQLSchema, graphql, graphql_sync
from graphql.utilities import print_schema
import logging
from sqlalchemy.orm import Session, DeclarativeBase
from sqlalchemy.ext.asyncio import AsyncSession
import time

from .errors import ConfigurationError
from .models import Table, Order
from .schema import build_gql_schema

log = logging.getLogger("alchemyql")


class AlchemyQL(ABC):
    """
    Alchemy QL Engine Definition.

    An Engine supports adding tables, building Graph QL schema and executing queries.
    """

    def __init__(self):
        """
        Initialize Alchemy QL Engine.
        """
        self.schema: GraphQLSchema | None = None
        self.tables: list[Table] = []
        self.is_async: bool

    def register(
        self,
        sqlalchemy_cls,
        graphql_name: str | None = None,
        description: str | None = None,
        include_fields: list[str] | None = None,
        exclude_fields: list[str] | None = None,
        filter_fields: list[str] | None = None,
        order_fields: list[str] | None = None,
        default_order: dict[str, Order] | None = None,
        pagination: bool = False,
        default_limit: int | None = None,
        max_limit: int | None = None,
    ):
        """
        Register a SQL Alchemy Table into your Alchemy QL engine.

        Options:
         - graphql_name - Name to give GraphQL type (defaults to tablename)
         - description - Description to give GraphQL type
         - include_fields - list of column names to expose
         - exclude_fields - list of column names not to expose
         - filter_fields - list of column names to allow filtering by
         - order_fields - list of column names to allow ordering by
         - default_order - column -> order map to be applied by default
         - pagination - whether to support pagination
         - default_limit - default max number of rows to return
         - max_limit - max limit to allow
        """

        table = Table(
            sqlalchemy_cls=sqlalchemy_cls,
            graphql_name=(graphql_name or sqlalchemy_cls.__tablename__).lower(),
            description=description,
            include_fields=include_fields,
            exclude_fields=exclude_fields or [],
            filter_fields=filter_fields or [],
            order_fields=order_fields or [],
            default_order=default_order,
            pagination=pagination,
            default_limit=default_limit,
            max_limit=max_limit,
        )

        # Checks the table is not already registerd
        if any(it.sqlalchemy_cls is sqlalchemy_cls for it in self.tables):
            raise ConfigurationError("Sqlalchemy Table is already registered")

        # Checks the name is not already in use
        if any(it.graphql_name == table.graphql_name for it in self.tables):
            raise ConfigurationError("Table with same Graph QL name already registerd")

        self.tables.append(table)
        log.debug(
            f"Registerd table {table.graphql_name}! (Now {len(self.tables)} tables registered!)"
        )

    def register_all_tables(self, base: type[DeclarativeBase]):
        """
        Register all tables under a DeclarativeBase.

        NOTE: This does not allow any customisations. It will default to use all columns.
        """
        for mapper in base.registry.mappers:
            self.register(mapper.class_)

    def build_schema(self):
        """
        Builds a Graph QL Schema using all registered SQL Alchemy Tables.

        NOTE: This must be run before the AlchemyQL engine can be used
        """
        start = time.perf_counter()

        self.schema = build_gql_schema(self.tables, self.is_async)

        log.debug(
            "Build schema complete! (Time taken: %.6f seconds)",
            time.perf_counter() - start,
        )

    def get_schema(self) -> str:
        """
        Returns the Graph QL Schema in a pretty print str format.
        """
        if not self.schema:
            raise ConfigurationError(
                "Schema is not setup yet. You must run 'build_schema()' first"
            )
        return print_schema(self.schema)


class AlchemyQLSync(AlchemyQL):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.is_async = False

    def execute_query(
        self,
        query: str,
        db_session: Session,
        variables: dict[str, Any] | None = None,
        operation: str | None = None,
    ) -> ExecutionResult:
        """
        Executes a Graph QL query on the Alchemy QL engine.
        """
        if not self.schema:
            raise ConfigurationError(
                "Schema is not setup yet. You must run 'build_schema()' first"
            )

        start = time.perf_counter()

        result = graphql_sync(
            self.schema,
            query,
            variable_values=variables,
            operation_name=operation,
            context_value={"session": db_session},
        )

        log.debug(
            "Query execution complete! (Time taken: %.6f seconds)",
            time.perf_counter() - start,
        )

        return result


class AlchemyQLAsync(AlchemyQL):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.is_async = True

    async def execute_query(
        self,
        query: str,
        db_session: AsyncSession,
        variables: dict[str, Any] | None = None,
        operation: str | None = None,
    ) -> ExecutionResult:
        """
        Executes a Graph QL query on the Alchemy QL engine.
        """
        if not self.schema:
            raise ConfigurationError(
                "Schema is not setup yet. You must run 'build_schema()' first"
            )

        start = time.perf_counter()

        result = await graphql(
            self.schema,
            query,
            variable_values=variables,
            operation_name=operation,
            context_value={"session": db_session},
        )

        log.debug(
            "Query execution complete! (Time taken: %.6f seconds)",
            time.perf_counter() - start,
        )

        return result



from datetime import date, datetime, time
from enum import Enum
from graphql import (
    GraphQLEnumType,
    GraphQLInputObjectType,
    GraphQLInputField,
    GraphQLInputType,
    GraphQLList,
)
from typing import cast

from .errors import ConfigurationError
from .scalars import (
    DateScalar,
    DateTimeScalar,
    TimeScalar,
    IntScalar,
    FloatScalar,
    BoolScalar,
    StringScalar,
)


IntFilter = cast(
    GraphQLInputObjectType,
    GraphQLInputObjectType(
        name="IntFilter",
        fields=lambda: {
            "eq": GraphQLInputField(IntScalar),
            "ne": GraphQLInputField(IntScalar),
            "gt": GraphQLInputField(IntScalar),
            "ge": GraphQLInputField(IntScalar),
            "lt": GraphQLInputField(IntScalar),
            "le": GraphQLInputField(IntScalar),
            "in": GraphQLInputField(GraphQLList(IntScalar)),
        },
    ),
)

FloatFilter = cast(
    GraphQLInputObjectType,
    GraphQLInputObjectType(
        name="FloatFilter",
        fields=lambda: {
            "eq": GraphQLInputField(FloatScalar),
            "ne": GraphQLInputField(FloatScalar),
            "gt": GraphQLInputField(FloatScalar),
            "ge": GraphQLInputField(FloatScalar),
            "lt": GraphQLInputField(FloatScalar),
            "le": GraphQLInputField(FloatScalar),
            "in": GraphQLInputField(GraphQLList(FloatScalar)),
        },
    ),
)

BoolFilter = cast(
    GraphQLInputObjectType,
    GraphQLInputObjectType(
        name="BoolFilter",
        fields=lambda: {
            "eq": GraphQLInputField(BoolScalar),
            "ne": GraphQLInputField(BoolScalar),
        },
    ),
)

StringFilter = cast(
    GraphQLInputObjectType,
    GraphQLInputObjectType(
        name="StringFilter",
        fields=lambda: {
            "eq": GraphQLInputField(StringScalar),
            "ne": GraphQLInputField(StringScalar),
            "contains": GraphQLInputField(StringScalar),
            "startswith": GraphQLInputField(StringScalar),
            "endswith": GraphQLInputField(StringScalar),
            "in": GraphQLInputField(GraphQLList(StringScalar)),
        },
    ),
)

DateTimeFilter = cast(
    GraphQLInputObjectType,
    GraphQLInputObjectType(
        name="DateTimeFilter",
        fields=lambda: {
            "eq": GraphQLInputField(DateTimeScalar),
            "ne": GraphQLInputField(DateTimeScalar),
            "gt": GraphQLInputField(DateTimeScalar),
            "ge": GraphQLInputField(DateTimeScalar),
            "lt": GraphQLInputField(DateTimeScalar),
            "le": GraphQLInputField(DateTimeScalar),
            "in": GraphQLInputField(GraphQLList(DateTimeScalar)),
        },
    ),
)

DateFilter = cast(
    GraphQLInputObjectType,
    GraphQLInputObjectType(
        name="DateFilter",
        fields=lambda: {
            "eq": GraphQLInputField(DateScalar),
            "ne": GraphQLInputField(DateScalar),
            "gt": GraphQLInputField(DateScalar),
            "ge": GraphQLInputField(DateScalar),
            "lt": GraphQLInputField(DateScalar),
            "le": GraphQLInputField(DateScalar),
            "in": GraphQLInputField(GraphQLList(DateScalar)),
        },
    ),
)

TimeFilter = cast(
    GraphQLInputObjectType,
    GraphQLInputObjectType(
        name="TimeFilter",
        fields=lambda: {
            "eq": GraphQLInputField(TimeScalar),
            "ne": GraphQLInputField(TimeScalar),
            "gt": GraphQLInputField(TimeScalar),
            "ge": GraphQLInputField(TimeScalar),
            "lt": GraphQLInputField(TimeScalar),
            "le": GraphQLInputField(TimeScalar),
            "in": GraphQLInputField(GraphQLList(TimeScalar)),
        },
    ),
)


def build_enum_filter(cls: GraphQLEnumType) -> GraphQLInputObjectType:
    return cast(
        GraphQLInputObjectType,
        GraphQLInputObjectType(
            name=f"{cls.name}_filter",
            fields=lambda: {
                "eq": GraphQLInputField(cls),
                "ne": GraphQLInputField(cls),
                "in": GraphQLInputField(GraphQLList(cls)),
            },
        ),
    )


def convert_to_filter(column, field: GraphQLInputType) -> GraphQLInputObjectType:
    py_type = column.type.python_type

    if py_type is int:
        return IntFilter
    elif py_type is str:
        return StringFilter
    elif py_type is bool:
        return BoolFilter
    elif py_type is float:
        return FloatFilter
    elif issubclass(py_type, Enum):
        return build_enum_filter(cast(GraphQLEnumType, field))
    elif py_type is datetime:
        return DateTimeFilter
    elif py_type is date:
        return DateFilter
    elif py_type is time:
        return TimeFilter
    raise ConfigurationError(
        "Filtering based on data types other than (int, str, bool, float, Enums, datetime, date, time) is not currently supported"
    )



class AlchemyQLError(Exception):
    """Base class for all exceptions raised by AlchemyQL."""

    pass


class ConfigurationError(AlchemyQLError):
    """Raised when the user provides invalid configuration."""

    pass


class QueryExecutionError(AlchemyQLError):
    """Raised when the query execution fails."""

    pass



from dataclasses import dataclass
from enum import auto, Enum


class Order(Enum):
    ASC = auto()
    DESC = auto()


@dataclass
class Table:
    # fmt: off

    # Table details
    sqlalchemy_cls  : type 
    graphql_name    : str
    description     : str | None

    # Fields Details
    include_fields  : list[str] | None
    exclude_fields  : list[str]

    # Filtering Details
    filter_fields    : list[str]

    # Ordering Details
    order_fields    : list[str]
    default_order   : dict[str, Order] | None

    # Pagination Details
    pagination      : bool
    default_limit   : int | None
    max_limit       : int | None

    # fmt: on



from enum import Enum
from sqlalchemy import select, desc, Select
from sqlalchemy.engine.row import RowMapping
from typing import Any

from .errors import QueryExecutionError
from .models import Table


def convert_enums(obj):
    """Recursively convert Enum objects inside dict/list/tuple/RowMapping."""

    # Enum → convert to string (name)
    if isinstance(obj, Enum):
        return obj.name

    # Dict/list, convert all
    elif isinstance(obj, (dict, RowMapping)):
        return {key: convert_enums(value) for key, value in obj.items()}
    elif isinstance(obj, (list, tuple)):
        return type(obj)(convert_enums(item) for item in obj)
    return obj


def build_sql_select_stmt(
    table,
    columns: list[str],
    filters: dict[str, Any] | None = None,
    offset: int | None = None,
    limit: int | None = None,
    order: dict[str, Any] | None = None,
) -> Select:
    """
    Build a SQLAlchemy Select statement based on GraphQL args.
    """
    # Step 1 - Build SELECT & FROM clauses
    cols = [getattr(table, col) for col in columns]
    stmt = select(*cols)

    # Step 2 - Build WHERE clause
    if filters:
        for col_name, operations in filters.items():
            column = getattr(table, col_name)
            for op, val in operations.items():
                if op == "eq":
                    stmt = stmt.where(column == val)
                elif op == "ne":
                    stmt = stmt.where(column != val)
                elif op == "lt":
                    stmt = stmt.where(column < val)
                elif op == "le":
                    stmt = stmt.where(column <= val)
                elif op == "gt":
                    stmt = stmt.where(column > val)
                elif op == "ge":
                    stmt = stmt.where(column >= val)
                elif op == "contains":
                    stmt = stmt.where(column.contains(val))
                elif op == "in":
                    stmt = stmt.where(column.in_(val))
                elif op == "startswith":
                    stmt = stmt.where(column.startswith(val))
                elif op == "endswith":
                    stmt = stmt.where(column.endswith(val))

    # Step 3 - Build pagination clauses (OFFSET, LIMIT)
    if offset is not None:
        stmt = stmt.offset(offset)

    if limit is not None:
        stmt = stmt.limit(limit)

    # Step 4 - Build ORDER BY clause
    if order:
        for col_name, direction in order.items():
            column = getattr(table, col_name)
            if str(direction).upper() == "DESC":
                column = desc(column)
            stmt = stmt.order_by(column)
    return stmt


def validations(table: Table, **kwargs):
    # Validate limit
    if limit := kwargs.get("limit"):
        if limit < 1 or (table.max_limit and limit > table.max_limit):
            raise QueryExecutionError(
                f"Provided Limit is out of bounds (Value: {limit}, Min: 1, Max: {table.max_limit})"
            )

    # Validate offset
    if offset := kwargs.get("offset"):
        if offset < 0:
            raise QueryExecutionError(
                f"Provided Offset is negative (Value: {offset}, Min: 0)"
            )


def build_async_resolver(table: Table):
    """
    Resolver function for Async queries.
    Returns a function that can be called at query execution to resolve query.
    """

    async def resolver(root, info, **kwargs):
        validations(table, **kwargs)

        db_session = info.context["session"]

        selection = info.field_nodes[0].selection_set

        query = build_sql_select_stmt(
            table=table.sqlalchemy_cls,
            columns=[f.name.value for f in selection.selections] if selection else [],
            filters=kwargs.get("filter", {}),
            offset=kwargs.get("offset", 0),
            limit=kwargs.get("limit", table.default_limit),
            order=kwargs.get("order", table.default_order),
        )

        res = await db_session.execute(query)

        return convert_enums(res.mappings().all())

    return resolver


def build_sync_resolver(table: Table):
    """
    Resolver function for Sync queries.
    Returns a function that can be called at query execution to resolve query.
    """

    def resolver(root, info, **kwargs):
        validations(table, **kwargs)

        db_session = info.context["session"]

        selection = info.field_nodes[0].selection_set

        query = build_sql_select_stmt(
            table=table.sqlalchemy_cls,
            columns=[f.name.value for f in selection.selections] if selection else [],
            filters=kwargs.get("filter", {}),
            offset=kwargs.get("offset", 0),
            limit=kwargs.get("limit", table.default_limit),
            order=kwargs.get("order", table.default_order),
        )

        res = db_session.execute(query)

        return convert_enums(res.mappings().all())

    return resolver



import base64
from datetime import date, datetime, time
from enum import Enum
from graphql import (
    GraphQLBoolean,
    GraphQLEnumType,
    GraphQLEnumValue,
    GraphQLFloat,
    GraphQLScalarType,
    GraphQLInputType,
    GraphQLInt,
    GraphQLString,
)
from typing import cast


IntScalar = cast(GraphQLInputType, GraphQLInt)
FloatScalar = cast(GraphQLInputType, GraphQLFloat)
BoolScalar = cast(GraphQLInputType, GraphQLBoolean)
StringScalar = cast(GraphQLInputType, GraphQLString)


BytesScalar = cast(
    GraphQLInputType,
    GraphQLScalarType(
        name="Bytes",
        serialize=lambda v: base64.b64encode(v).decode("ascii")
        if isinstance(v, (bytes, bytearray))
        else None,
        parse_value=lambda v: base64.b64decode(v.encode("ascii")),
    ),
)

JSONScalar = cast(
    GraphQLInputType,
    GraphQLScalarType(
        name="JSON",
        serialize=lambda v: v,
        parse_value=lambda v: v,
    ),
)

TimeScalar = cast(
    GraphQLInputType,
    GraphQLScalarType(
        name="Time",
        serialize=lambda v: v.isoformat() if isinstance(v, time) else None,
        parse_value=lambda v: time.fromisoformat(v),
    ),
)

DateTimeScalar = cast(
    GraphQLInputType,
    GraphQLScalarType(
        name="DateTime",
        serialize=lambda v: v.isoformat() if isinstance(v, datetime) else None,
        parse_value=lambda v: datetime.fromisoformat(v),
    ),
)

DateScalar = cast(
    GraphQLInputType,
    GraphQLScalarType(
        name="Date",
        serialize=lambda v: v.isoformat() if isinstance(v, date) else None,
        parse_value=lambda v: date.fromisoformat(v),
    ),
)


def build_enum_scalar(cls: type[Enum]) -> GraphQLInputType:
    return cast(
        GraphQLInputType,
        GraphQLEnumType(
            name=cls.__name__,
            values={mem.name: GraphQLEnumValue(mem.name) for mem in cls},
        ),
    )


class Order(Enum):
    ASC = "ASC"
    DESC = "DESC"


OrderingEnumScalar = build_enum_scalar(Order)


def convert_to_scalar(column) -> GraphQLInputType:
    py_type = column.type.python_type

    if py_type is int:
        return IntScalar
    elif py_type is bool:
        return BoolScalar
    elif py_type is float:
        return FloatScalar
    elif issubclass(py_type, Enum):
        return build_enum_scalar(py_type)
    elif py_type is datetime:
        return DateTimeScalar
    elif py_type is date:
        return DateScalar
    elif py_type is time:
        return TimeScalar
    elif py_type is dict or py_type is list:
        return JSONScalar
    elif py_type is bytearray or py_type is bytes:
        return BytesScalar
    return StringScalar
